@@ -147,6 +147,7 @@ int htp_is_space(int c) {
  * @return Method number of M_UNKNOWN
  */
 int htp_convert_method_to_number(bstr *method) {
+    if (method == NULL) return M_UNKNOWN;
     // TODO Optimize using parallel matching, or something
     if (bstr_cmpc(method, "GET") == 0) return M_GET;
     if (bstr_cmpc(method, "PUT") == 0) return M_PUT;
@@ -306,7 +307,7 @@ void htp_log(htp_connp_t *connp, const char *file, int line, int level, int code
     va_start(args, fmt);
 
     int r = vsnprintf(buf, 1023, fmt, args);
-    
+
     va_end(args);
 
     if (r < 0) {
@@ -337,7 +338,7 @@ void htp_log(htp_connp_t *connp, const char *file, int line, int level, int code
         connp->last_error = log;
     }
 
-    hook_run_all(connp->cfg->hook_log, log);   
+    hook_run_all(connp->cfg->hook_log, log);
 }
 
 /**
@@ -446,6 +447,8 @@ int htp_parse_authority(htp_connp_t *connp, bstr *authority, htp_uri_t **uri) {
  * @return HTP_ERROR on memory allocation failure, HTP_OK otherwise
  */
 int htp_parse_uri(bstr *input, htp_uri_t **uri) {
+    if (input == NULL)
+        return HTP_ERROR;
     char *data = bstr_ptr(input);
     size_t len = bstr_len(input);
     size_t start, pos;
@@ -656,6 +659,9 @@ uint8_t bestfit_codepoint(htp_cfg_t *cfg, uint32_t codepoint) {
  * @param path
  */
 void htp_utf8_decode_path_inplace(htp_cfg_t *cfg, htp_tx_t *tx, bstr *path) {
+    if (path == NULL)
+        return;
+
     uint8_t *data = (unsigned char *) bstr_ptr(path);
     size_t len = bstr_len(path);
     size_t rpos = 0;
@@ -920,6 +926,9 @@ int decode_u_encoding(htp_cfg_t *cfg, htp_tx_t *tx, unsigned char *data) {
  * @param path
  */
 int htp_decode_path_inplace(htp_cfg_t *cfg, htp_tx_t *tx, bstr *path) {
+    if (path == NULL)
+        return -1;
+
     unsigned char *data = (unsigned char *) bstr_ptr(path);
     if (data == NULL) {
         return -1;
@@ -1192,17 +1201,23 @@ int htp_normalize_parsed_uri(htp_connp_t *connp, htp_uri_t *incomplete, htp_uri_
     if (incomplete->scheme != NULL) {
         // Duplicate and convert to lowercase
         normalized->scheme = bstr_dup_lower(incomplete->scheme);
+        if (normalized->scheme == NULL)
+            return HTP_ERROR;
     }
 
     // Username
     if (incomplete->username != NULL) {
         normalized->username = bstr_strdup(incomplete->username);
+        if (normalized->username == NULL)
+            return HTP_ERROR;
         htp_uriencoding_normalize_inplace(normalized->username);
     }
 
     // Password
     if (incomplete->password != NULL) {
         normalized->password = bstr_strdup(incomplete->password);
+        if (normalized->password == NULL)
+            return HTP_ERROR;
         htp_uriencoding_normalize_inplace(normalized->password);
     }
 
@@ -1211,6 +1226,8 @@ int htp_normalize_parsed_uri(htp_connp_t *connp, htp_uri_t *incomplete, htp_uri_
         // We know that incomplete->hostname does not contain
         // port information, so no need to check for it here
         normalized->hostname = bstr_strdup(incomplete->hostname);
+        if (normalized->hostname == NULL)
+            return HTP_ERROR;
         htp_uriencoding_normalize_inplace(normalized->hostname);
         htp_normalize_hostname_inplace(normalized->hostname);
     }
@@ -1228,22 +1245,25 @@ int htp_normalize_parsed_uri(htp_connp_t *connp, htp_uri_t *incomplete, htp_uri_
     if (incomplete->path != NULL) {
         // Make a copy of the path, on which we can work on
         normalized->path = bstr_strdup(incomplete->path);
+        if (normalized->path != NULL) {
+            // Decode URL-encoded (and %u-encoded) characters, as well as lowercase,
+            // compress separators and convert backslashes.
+            htp_decode_path_inplace(connp->cfg, connp->in_tx, normalized->path);
 
-        // Decode URL-encoded (and %u-encoded) characters, as well as lowercase,
-        // compress separators and convert backslashes.
-        htp_decode_path_inplace(connp->cfg, connp->in_tx, normalized->path);
+            // Handle UTF-8 in path
+            if (connp->cfg->path_convert_utf8) {
+                // Decode Unicode characters into a single-byte stream, using best-fit mapping
+                htp_utf8_decode_path_inplace(connp->cfg, connp->in_tx, normalized->path);
+            } else {
+                // Only validate path as a UTF-8 stream
+                htp_utf8_validate_path(connp->in_tx, normalized->path);
+            }
 
-        // Handle UTF-8 in path
-        if (connp->cfg->path_convert_utf8) {
-            // Decode Unicode characters into a single-byte stream, using best-fit mapping
-            htp_utf8_decode_path_inplace(connp->cfg, connp->in_tx, normalized->path);
+            // RFC normalization
+            htp_normalize_uri_path_inplace(normalized->path);
         } else {
-            // Only validate path as a UTF-8 stream
-            htp_utf8_validate_path(connp->in_tx, normalized->path);
+            return HTP_ERROR;
         }
-
-        // RFC normalization
-        htp_normalize_uri_path_inplace(normalized->path);
     }
 
     // Query
@@ -1251,11 +1271,15 @@ int htp_normalize_parsed_uri(htp_connp_t *connp, htp_uri_t *incomplete, htp_uri_
         // We cannot URL-decode the query string here; it needs to be
         // parsed into individual key-value pairs first.
         normalized->query = bstr_strdup(incomplete->query);
+        if (normalized->query == NULL)
+            return HTP_ERROR;
     }
 
     // Fragment
     if (incomplete->fragment != NULL) {
         normalized->fragment = bstr_strdup(incomplete->fragment);
+        if (normalized->fragment == NULL)
+            return HTP_ERROR;
         htp_uriencoding_normalize_inplace(normalized->fragment);
     }
 
@@ -1270,6 +1294,8 @@ int htp_normalize_parsed_uri(htp_connp_t *connp, htp_uri_t *incomplete, htp_uri_
  * @return normalized hostnanme
  */
 bstr *htp_normalize_hostname_inplace(bstr *hostname) {
+    if (hostname == NULL)
+        return NULL;
     bstr_tolowercase(hostname);
 
     char *data = bstr_ptr(hostname);
@@ -1294,6 +1320,8 @@ bstr *htp_normalize_hostname_inplace(bstr *hostname) {
  * @param hostname
  */
 void htp_replace_hostname(htp_connp_t *connp, htp_uri_t *parsed_uri, bstr *hostname) {
+    if (hostname == NULL)
+        return;
     int colon = bstr_chr(hostname, ':');
     if (colon == -1) {
         // Hostname alone
@@ -1348,6 +1376,7 @@ int htp_is_uri_unreserved(unsigned char c) {
  * @param s
  */
 void htp_uriencoding_normalize_inplace(bstr *s) {
+    if (s == NULL) return;
     unsigned char *data = (unsigned char *) bstr_ptr(s);
     size_t len = bstr_len(s);
 
@@ -1474,6 +1503,7 @@ int htp_prenormalize_uri_path_inplace(bstr *s, int *flags, int case_insensitive,
  * @param s
  */
 void htp_normalize_uri_path_inplace(bstr *s) {
+    if (s == NULL) return;
     char *data = bstr_ptr(s);
     size_t len = bstr_len(s);
 
@@ -1710,7 +1740,7 @@ char *htp_tx_progress_as_string(htp_tx_t *tx) {
             return "DONE";
     }
 
-    return "UNKOWN";
+    return "UNKNOWN";
 }
 
 bstr *htp_unparse_uri_noencode(htp_uri_t *uri) {
@@ -1830,6 +1860,10 @@ bstr *htp_tx_generate_request_headers_raw(htp_tx_t *tx) {
     for (i = 0; i < list_size(tx->request_header_lines); i++) {
         htp_header_line_t *hl = list_get(tx->request_header_lines, i);
         len += bstr_len(hl->line);
+        if (hl->terminators)
+            len += bstr_len(hl->terminators);
+        else
+            len += 2; // 0d 0a
     }
 
     request_headers_raw = bstr_alloc(len);
@@ -1841,6 +1875,10 @@ bstr *htp_tx_generate_request_headers_raw(htp_tx_t *tx) {
     for (i = 0; i < list_size(tx->request_header_lines); i++) {
         htp_header_line_t *hl = list_get(tx->request_header_lines, i);
         bstr_add_str_noex(request_headers_raw, hl->line);
+        if (hl->terminators)
+            bstr_add_str_noex(request_headers_raw, hl->terminators);
+        else
+            bstr_add_cstr_noex(request_headers_raw, "\r\n");
     }
 
     return request_headers_raw;
@@ -1878,3 +1916,73 @@ bstr *htp_tx_get_request_headers_raw(htp_tx_t *tx) {
 
     return tx->request_headers_raw;
 }
+
+/**
+ * Construct a bstr that contains the raw response headers.
+ *
+ * @param tx
+ * @return
+ */
+bstr *htp_tx_generate_response_headers_raw(htp_tx_t *tx) {
+    bstr *response_headers_raw = NULL;
+    size_t i, len = 0;
+
+    for (i = 0; i < list_size(tx->response_header_lines); i++) {
+        htp_header_line_t *hl = list_get(tx->response_header_lines, i);
+        len += bstr_len(hl->line);
+        if (hl->terminators)
+            len += bstr_len(hl->terminators);
+        else
+            len += 2; // 0d 0a
+    }
+
+    response_headers_raw = bstr_alloc(len);
+    if (response_headers_raw == NULL) {
+        htp_log(tx->connp, HTP_LOG_MARK, HTP_LOG_ERROR, 0, "Failed to allocate bstring of %d bytes", len);
+        return NULL;
+    }
+
+    for (i = 0; i < list_size(tx->response_header_lines); i++) {
+        htp_header_line_t *hl = list_get(tx->response_header_lines, i);
+        bstr_add_str_noex(response_headers_raw, hl->line);
+        if (hl->terminators)
+            bstr_add_str_noex(response_headers_raw, hl->terminators);
+        else
+            bstr_add_cstr_noex(response_headers_raw, "\r\n");
+    }
+
+    return response_headers_raw;
+}
+
+/**
+ * Get a bstr that contains the raw response headers. This method will always
+ * return an up-to-date buffer, containing the last known headers. Thus, if
+ * it is called once after RESPONSE_HEADERS phase it will return one buffer, but
+ * it may return a different buffer if called after RESPONSE_TRAILERS phase (but
+ * only if the response actually contains trailer headers). Do not retain the
+ * bstr pointer, as the buffer may change. If there are no changes to the
+ * response header structure, only one buffer will be contstructed and used. (Multiple
+ * invocations of this method will not cause multiple buffers to be created.)
+ *
+ * @param tx
+ * @return
+ */
+bstr *htp_tx_get_response_headers_raw(htp_tx_t *tx) {
+    // Check that we are not called too early
+    if (tx->progress < TX_PROGRESS_RES_HEADERS) return NULL;
+
+    if (tx->response_headers_raw == NULL) {
+        tx->response_headers_raw = htp_tx_generate_response_headers_raw(tx);
+        tx->response_headers_raw_lines = list_size(tx->response_header_lines);
+    } else {
+        // Check that the buffer we have is not obsolete
+        if (tx->response_headers_raw_lines < list_size(tx->response_header_lines)) {
+            // Rebuild raw buffer
+            bstr_free(tx->response_headers_raw);
+            tx->response_headers_raw = htp_tx_generate_response_headers_raw(tx);
+            tx->response_headers_raw_lines = list_size(tx->response_header_lines);
+        }
+    }
+
+    return tx->response_headers_raw;
+}
