@@ -1,3 +1,16 @@
+/*
+ * LibHTP (http://www.libhtp.org)
+ * Copyright 2009,2010 Ivan Ristic <ivanr@webkreator.com>
+ *
+ * LibHTP is an open source product, released under terms of the General Public Licence
+ * version 2 (GPLv2). Please refer to the file LICENSE, which contains the complete text
+ * of the license.
+ *
+ * In addition, there is a special exception that allows LibHTP to be freely
+ * used with any OSI-approved open source licence. Please refer to the file
+ * LIBHTP_LICENSING_EXCEPTION for the full text of the exception.
+ *
+ */
 
 #include <dirent.h>
 #include <stdio.h>
@@ -18,10 +31,13 @@ char *home = NULL;
 int test_get(htp_cfg_t *cfg) {
     htp_connp_t *connp = NULL;
 
-    test_run(home, "01-get.t", cfg, &connp);
-    if (connp == NULL) return -1;
+    int rc = test_run(home, "01-get.t", cfg, &connp);
+    if (rc < 0) {
+        if (connp != NULL) htp_connp_destroy_all(connp);
+        return -1;
+    }
 
-    htp_connp_destroy_all(connp);
+    htp_connp_destroy_all(connp);    
 
     return 1;
 }
@@ -32,8 +48,11 @@ int test_get(htp_cfg_t *cfg) {
 int test_post_urlencoded_chunked(htp_cfg_t *cfg) {
     htp_connp_t *connp = NULL;
 
-    test_run(home, "04-post-urlencoded-chunked.t", cfg, &connp);
-    if (connp == NULL) return -1;
+    int rc = test_run(home, "04-post-urlencoded-chunked.t", cfg, &connp);
+    if (rc < 0) {
+        if (connp != NULL) htp_connp_destroy_all(connp);
+        return -1;
+    }
 
     htp_tx_t *tx = list_get(connp->conn->transactions, 0);
 
@@ -48,6 +67,9 @@ int test_post_urlencoded_chunked(htp_cfg_t *cfg) {
         free(key);
     }
 
+    bstr *raw = htp_tx_get_request_headers_raw(tx);
+    fprint_raw_data(stdout, "REQUEST HEADERS RAW 2", bstr_ptr(raw), bstr_len(raw));
+
     htp_connp_destroy_all(connp);
 
     return 1;
@@ -59,8 +81,11 @@ int test_post_urlencoded_chunked(htp_cfg_t *cfg) {
 int test_post_urlencoded(htp_cfg_t *cfg) {
     htp_connp_t *connp = NULL;
 
-    test_run(home, "03-post-urlencoded.t", cfg, &connp);
-    if (connp == NULL) return -1;
+    int rc = test_run(home, "03-post-urlencoded.t", cfg, &connp);
+    if (rc < 0) {
+        if (connp != NULL) htp_connp_destroy_all(connp);
+        return -1;
+    }
 
     htp_connp_destroy_all(connp);
 
@@ -73,8 +98,11 @@ int test_post_urlencoded(htp_cfg_t *cfg) {
 int test_apache_header_parsing(htp_cfg_t *cfg) {
     htp_connp_t *connp = NULL;
 
-    test_run(home, "02-header-test-apache2.t", cfg, &connp);
-    if (connp == NULL) return -1;
+    int rc = test_run(home, "02-header-test-apache2.t", cfg, &connp);
+    if (rc < 0) {
+        if (connp != NULL) htp_connp_destroy_all(connp);
+        return -1;
+    }
 
     htp_tx_t *tx = list_get(connp->conn->transactions, 0);
     if (tx == NULL) return -1;
@@ -218,8 +246,11 @@ int test_apache_header_parsing(htp_cfg_t *cfg) {
 int test_expect(htp_cfg_t *cfg) {
     htp_connp_t *connp = NULL;
 
-    test_run(home, "05-expect.t", cfg, &connp);
-    if (connp == NULL) return -1;
+    int rc = test_run(home, "05-expect.t", cfg, &connp);
+    if (rc < 0) {
+        if (connp != NULL) htp_connp_destroy_all(connp);
+        return -1;
+    }
 
     htp_connp_destroy_all(connp);
 
@@ -232,8 +263,11 @@ int test_expect(htp_cfg_t *cfg) {
 int test_uri_normal(htp_cfg_t *cfg) {
     htp_connp_t *connp = NULL;
 
-    test_run(home, "06-uri-normal.t", cfg, &connp);
-    if (connp == NULL) return -1;
+    int rc = test_run(home, "06-uri-normal.t", cfg, &connp);
+    if (rc < 0) {
+        if (connp != NULL) htp_connp_destroy_all(connp);
+        return -1;
+    }
 
     htp_connp_destroy_all(connp);
 
@@ -246,8 +280,11 @@ int test_uri_normal(htp_cfg_t *cfg) {
 int test_pipelined_connection(htp_cfg_t *cfg) {
     htp_connp_t *connp = NULL;
 
-    test_run(home, "07-pipelined-connection.t", cfg, &connp);
-    if (connp == NULL) return -1;
+    int rc = test_run(home, "07-pipelined-connection.t", cfg, &connp);
+    if (rc < 0) {
+        if (connp != NULL) htp_connp_destroy_all(connp);
+        return -1;
+    }
 
     if (list_size(connp->conn->transactions) != 2) {
         printf("Expected 2 transactions but found %i.", list_size(connp->conn->transactions));
@@ -270,8 +307,11 @@ int test_pipelined_connection(htp_cfg_t *cfg) {
 int test_not_pipelined_connection(htp_cfg_t *cfg) {
     htp_connp_t *connp = NULL;
 
-    test_run(home, "08-not-pipelined-connection.t", cfg, &connp);
-    if (connp == NULL) return -1;
+    int rc = test_run(home, "08-not-pipelined-connection.t", cfg, &connp);
+    if (rc < 0) {
+        if (connp != NULL) htp_connp_destroy_all(connp);
+        return -1;
+    }
 
     if (list_size(connp->conn->transactions) != 2) {
         printf("Expected 2 transactions but found %i.", list_size(connp->conn->transactions));
@@ -301,8 +341,11 @@ int test_not_pipelined_connection(htp_cfg_t *cfg) {
 int test_multi_packet_request_head(htp_cfg_t *cfg) {
     htp_connp_t *connp = NULL;
 
-    test_run(home, "09-multi-packet-request-head.t", cfg, &connp);
-    if (connp == NULL) return -1;
+    int rc = test_run(home, "09-multi-packet-request-head.t", cfg, &connp);
+    if (rc < 0) {
+        if (connp != NULL) htp_connp_destroy_all(connp);
+        return -1;
+    }
 
     if (list_size(connp->conn->transactions) != 1) {
         printf("Expected 1 transaction but found %i.", list_size(connp->conn->transactions));
@@ -324,8 +367,11 @@ int test_multi_packet_request_head(htp_cfg_t *cfg) {
 int test_misc(htp_cfg_t *cfg) {
     htp_connp_t *connp = NULL;
 
-    test_run(home, "misc.t", cfg, &connp);
-    if (connp == NULL) return -1;
+    int rc = test_run(home, "misc.t", cfg, &connp);
+    if (rc < 0) {
+        if (connp != NULL) htp_connp_destroy_all(connp);
+        return -1;
+    }
 
     if (list_size(connp->conn->transactions) == 0) {
         printf("Expected at least one transaction");
@@ -347,8 +393,11 @@ int test_misc(htp_cfg_t *cfg) {
 int test_host_in_headers(htp_cfg_t *cfg) {
     htp_connp_t *connp = NULL;
 
-    test_run(home, "10-host-in-headers.t", cfg, &connp);
-    if (connp == NULL) return -1;
+    int rc = test_run(home, "10-host-in-headers.t", cfg, &connp);
+    if (rc < 0) {
+        if (connp != NULL) htp_connp_destroy_all(connp);
+        return -1;
+    }
 
     if (list_size(connp->conn->transactions) != 4) {
         printf("Expected 4 transactions but found %i.", list_size(connp->conn->transactions));
@@ -388,8 +437,11 @@ int test_host_in_headers(htp_cfg_t *cfg) {
 int test_response_stream_closure(htp_cfg_t *cfg) {
     htp_connp_t *connp = NULL;
 
-    test_run(home, "11-response-stream-closure.t", cfg, &connp);
-    if (connp == NULL) return -1;
+    int rc = test_run(home, "11-response-stream-closure.t", cfg, &connp);
+    if (rc < 0) {
+        if (connp != NULL) htp_connp_destroy_all(connp);
+        return -1;
+    }
 
     if (list_size(connp->conn->transactions) == 0) {
         printf("Expected at least one transaction");
@@ -411,8 +463,11 @@ int test_response_stream_closure(htp_cfg_t *cfg) {
 int test_connect(htp_cfg_t *cfg) {
     htp_connp_t *connp = NULL;
 
-    test_run(home, "12-connect-request.t", cfg, &connp);
-    if (connp == NULL) return -1;
+    int rc = test_run(home, "12-connect-request.t", cfg, &connp);
+    if (rc < 0) {
+        if (connp != NULL) htp_connp_destroy_all(connp);
+        return -1;
+    }
 
     if (list_size(connp->conn->transactions) == 0) {
         printf("Expected at least one transaction");
@@ -435,11 +490,66 @@ int test_connect(htp_cfg_t *cfg) {
     return 1;
 }
 
+int test_connect_complete(htp_cfg_t *cfg) {
+    htp_connp_t *connp = NULL;
+
+    int rc = test_run(home, "15-connect-complete.t", cfg, &connp);
+    if (rc < 0) {
+        if (connp != NULL) htp_connp_destroy_all(connp);
+        return -1;
+    }
+
+    if (list_size(connp->conn->transactions) == 0) {
+        printf("Expected at least one transaction");
+        return -1;
+    }
+
+    htp_tx_t *tx = list_get(connp->conn->transactions, 0);
+
+    if (tx->progress != TX_PROGRESS_DONE) {
+        printf("Expected the only transaction to be complete (but got %i).", tx->progress);
+        return -1;
+    }   
+
+    htp_connp_destroy_all(connp);
+
+    return 1;
+}
+
+int test_connect_extra(htp_cfg_t *cfg) {
+    htp_connp_t *connp = NULL;
+
+    int rc = test_run(home, "16-connect-extra.t", cfg, &connp);
+    if (rc < 0) {
+        if (connp != NULL) htp_connp_destroy_all(connp);
+        return -1;
+    }
+
+    if (list_size(connp->conn->transactions) == 0) {
+        printf("Expected at least one transaction");
+        return -1;
+    }
+
+    htp_tx_t *tx = list_get(connp->conn->transactions, 0);
+
+    if (tx->progress != TX_PROGRESS_DONE) {
+        printf("Expected the only transaction to be complete (but got %i).", tx->progress);
+        return -1;
+    }
+
+    htp_connp_destroy_all(connp);
+
+    return 1;
+}
+
 int test_compressed_response_gzip_ct(htp_cfg_t *cfg) {
     htp_connp_t *connp = NULL;
 
-    test_run(home, "13-compressed-response-gzip-ct.t", cfg, &connp);
-    if (connp == NULL) return -1;
+    int rc = test_run(home, "13-compressed-response-gzip-ct.t", cfg, &connp);
+    if (rc < 0) {
+        if (connp != NULL) htp_connp_destroy_all(connp);
+        return -1;
+    }
 
     if (list_size(connp->conn->transactions) == 0) {
         printf("Expected at least one transaction");
@@ -461,8 +571,11 @@ int test_compressed_response_gzip_ct(htp_cfg_t *cfg) {
 int test_compressed_response_gzip_chunked(htp_cfg_t *cfg) {
     htp_connp_t *connp = NULL;
 
-    test_run(home, "14-compressed-response-gzip-chunked.t", cfg, &connp);
-    if (connp == NULL) return -1;
+    int rc = test_run(home, "14-compressed-response-gzip-chunked.t", cfg, &connp);
+    if (rc < 0) {
+        if (connp != NULL) htp_connp_destroy_all(connp);
+        return -1;
+    }
 
     if (list_size(connp->conn->transactions) == 0) {
         printf("Expected at least one transaction");
@@ -622,7 +735,7 @@ int main_dir(int argc, char** argv) {
     htp_config_register_log(cfg, callback_log);
     htp_config_register_response(cfg, callback_response_destroy);
     
-    run_directory("C:\\http_traces\\run-nikto-faulty", cfg);
+    run_directory("C:\\http_traces\\run5", cfg);
     //run_directory("/home/ivanr/work/traces/run3/", cfg);
     
     htp_config_destroy(cfg);
@@ -704,23 +817,29 @@ int main(int argc, char** argv) {
     htp_config_register_response(cfg, callback_response);
 
     htp_config_register_log(cfg, callback_log);
+
+    htp_config_set_generate_request_uri_normalized(cfg, 1);
     
     RUN_TEST(test_get, cfg);
-    RUN_TEST(test_apache_header_parsing, cfg);
-    RUN_TEST(test_post_urlencoded, cfg);
-    RUN_TEST(test_post_urlencoded_chunked, cfg);
-    RUN_TEST(test_expect, cfg);
-    RUN_TEST(test_uri_normal, cfg);
-    RUN_TEST(test_pipelined_connection, cfg);
-    RUN_TEST(test_not_pipelined_connection, cfg);
-    RUN_TEST(test_multi_packet_request_head, cfg);
-    RUN_TEST(test_response_stream_closure, cfg);
-    RUN_TEST(test_host_in_headers, cfg);
-    RUN_TEST(test_connect, cfg);
-    RUN_TEST(test_compressed_response_gzip_ct, cfg);
-    RUN_TEST(test_compressed_response_gzip_chunked, cfg);
+    //RUN_TEST(test_apache_header_parsing, cfg);
+    //RUN_TEST(test_post_urlencoded, cfg);
+    //RUN_TEST(test_post_urlencoded_chunked, cfg);
+    //RUN_TEST(test_expect, cfg);
+    //RUN_TEST(test_uri_normal, cfg);
+    //RUN_TEST(test_pipelined_connection, cfg);
+    //RUN_TEST(test_not_pipelined_connection, cfg);
+    //RUN_TEST(test_multi_packet_request_head, cfg);
+    //RUN_TEST(test_response_stream_closure, cfg);
+    //RUN_TEST(test_host_in_headers, cfg);
+    //RUN_TEST(test_compressed_response_gzip_ct, cfg);
+    //RUN_TEST(test_compressed_response_gzip_chunked, cfg);
+    
+    //RUN_TEST(test_connect, cfg);
+    //RUN_TEST(test_connect_complete, cfg);
+    //RUN_TEST(test_connect_extra, cfg);    
 
     //RUN_TEST(test_misc, cfg);
+    //RUN_TEST(test_post_urlencoded_chunked, cfg);
 
     printf("Tests: %i\n", tests);
     printf("Failures: %i\n", failures);
@@ -915,6 +1034,7 @@ int main_utf8_decoder_tests(int argc, char** argv) {
     bstr_free(expected); \
     bstr_free(input);
 
+
 int main_path_tests(int argc, char** argv) {
     htp_cfg_t *cfg = NULL;
     htp_tx_t *tx = NULL;
